// (C) 2017-2020 by folkert van heusden, released under AGPL v3.0
#include <signal.h>
#include <stdio.h>
#include <thread>
#include <unistd.h>
#include <vector>

#include <curl/curl.h>

#include "cfg.h"
#include "log.h"
#include "target.h"
#include "error.h"

void version()
{
	printf(NAME " " VERSION "\n");
	printf("(C) 2017-2020 by Folkert van Heusden\n\n");
}

void help()
{
	printf("-c x   select configuration file\n");
	printf("-f     fork into the background.\n");
	printf("       when not forking, then the program can be stopped by pressing enter.\n");
	printf("-p x   file to write PID to\n");
	printf("-v     enable verbose mode\n");
	printf("-V     show version & exit\n");
	printf("-h     this help\n");
}

int main(int argc, char *argv[])
{
	const char *pid_file = NULL;
	bool do_fork = false, verbose = false;
	int ll = LL_INFO;

	configuration_t cfg;
	cfg.lock.lock();

	cfg.cfg_file = "constatus.cfg";

	int c = -1;
	while((c = getopt(argc, argv, "c:p:fhvV")) != -1) {
		switch(c) {
			case 'c':
				cfg.cfg_file = optarg;
				break;

			case 'p':
				pid_file = optarg;
				break;

			case 'f':
				do_fork = true;
				break;

			case 'h':
				help();
				return 0;

			case 'v':
				verbose = true;
				ll++;
				break;

			case 'V':
				version();
				return 0;

			default:
				help();
				return 1;
		}
	}

	cfg.lock.unlock();

	signal(SIGCHLD, SIG_IGN);
	signal(SIGPIPE, SIG_IGN);

	curl_global_init(CURL_GLOBAL_ALL);

	load_configuration(&cfg, verbose, ll);

	if (pid_file && !do_fork)
		log(LL_WARNING, "Will not write a PID file when not forking");

	if (do_fork) {
		if (daemon(0, 0) == -1)
			error_exit(true, "daemon() failed");

		if (pid_file) {
			FILE *fh = fopen(pid_file, "w");
			if (!fh)
				error_exit(true, "Cannot create %s", pid_file);


			fprintf(fh, "%d", getpid());
			fclose(fh);
		}

		for(;;)
			sleep(86400);
	}

	log(LL_INFO, "Starting threads");

	struct timeval ets;
	gettimeofday(&ets, nullptr);

	for(instance_t * i : cfg.instances) {
		log(LL_DEBUG, "Instance %s", i->name.c_str());

		unsigned long int event_nr = 0;

		for(interface *t : i -> interfaces) {
			log(LL_DEBUG, "Starting %s", t->get_id().c_str());

			if (t->get_class_type() == CT_SOURCE)
				event_nr = t->get_db() -> register_event(t->get_id(), EVT_GENERIC, "regular startup", &ets);

			if (t->get_class_type() == CT_TARGET)
				((target *)t)->start(nullptr, event_nr);
			else
				t->start();
		}
	}

	for(interface *t : cfg.global_http_servers)
		t -> start();

	if (cfg.clnr) {
		log(LL_INFO, "Starting cleaner");
		cfg.clnr -> start();
	}

	log(LL_INFO, "System started");

	if (do_fork) {
		for(;;)
			sleep(86400);
	}
	else {
		getchar();
	}

	log(LL_INFO, "Terminating");

	cfg.lock.lock();

	for(interface *t : cfg.global_http_servers)
		t -> stop();

	for(instance_t * i : cfg.instances) {
		for(interface *t : i -> interfaces)
			t -> stop();
	}

	if (cfg.clnr)
		cfg.clnr -> stop();

	log(LL_INFO, "Waiting for threads to terminate");

	for(interface *t : cfg.global_http_servers)
		delete t;

	for(instance_t * i : cfg.instances) {
		for(size_t l=0; l<i->interfaces.size();) {
			interface *t = i->interfaces.at(l);

			if (t->get_class_type() != CT_SOURCE && t->get_class_type() != CT_VIEW) {
				i->interfaces.erase(i->interfaces.begin() + l);
				delete t;
			}
			else {
				l++;
			}
		}
	}

	for(instance_t * i : cfg.instances) {
		for(size_t l=0; l<i->interfaces.size();) {
			interface *t = i->interfaces.at(l);

			if (t->get_class_type() == CT_VIEW) {
				i->interfaces.erase(i->interfaces.begin() + l);
				delete t;
			}
			else {
				l++;
			}
		}
	}

	for(instance_t * i : cfg.instances) {
		for(size_t l=0; l<i->interfaces.size();) {
			interface *t = i->interfaces.at(l);

			if (t->get_class_type() == CT_SOURCE) {
				i->interfaces.erase(i->interfaces.begin() + l);
				delete t;
			}
			else {
				l++;
			}
		}

		delete i;
	}

	delete cfg.r;

	delete cfg.clnr;

	// FIXME	if (et) {
	// FIXME		dlclose(et -> library);
	// FIXME		delete et;
	// FIXME	}

	curl_global_cleanup();

	log(LL_INFO, "Bye bye");

	return 0;
}
